from model import Model, _dict_to_matrix, _dict_to_vector, _vector_to_dict
from fit import fit
from scipy.sparse import bsr_matrix
from numpy import dot, vdot, inf
from math import sqrt, exp, log
from vptree import VPTree
from sortedcollection import SortedCollection
import random
import numpy
import sys

class MetricError:
    """Raised when a non-positive-definite metric is encountered.

    """
    pass

class ScanDataError:
    """Raised for queries from empty ScanData objects.

    """
    pass

class _ScanItem:
    def __init__(self,
                 pvec=None,
                 xvec=None,
                 chisq=None,
                 cons=None,
                 nndist=None,
                 fitness=None):
        self.pvec    = pvec
        self.xvec    = xvec
        self.chisq   = chisq
        self.cons    = cons
        self.nndist  = nndist
        self.fitness = fitness

def _multichoice(seq, n):
    r = range(len(seq))
    res = []
    for i in range(n):
        j = int(random.uniform(0, len(r)))
        res.append(seq[r[j]])
        del r[j]
    return tuple(res)

class Fitter:
    """Sample a model's parameter space and minimise the chi-square.

    Args:
      model (Model instance): The model to fit.
      tries (int, optional): The number of local minimizations to try.
        Defaults to 1.
      retries (int, optional): Maximum number of retries (with a different
        starting point) after a failed minimization. Defaults to 0.
      cvmargin (float, optional): The acceptable amount of constraint
        violation when selecting starting points for the optimization. The
        sum of squares of the constraint functions must be smaller than
        `cvmargin`. Defaults to 10.
      maxcv (float, optional): Maximal value of the constraint violation (sum of
        squares of all violated constraint functions) for which the optimization
        is still considered successful. Defaults to ``1e-4``.
      solver (str, optional): Identifies the optimization algorithm to use.
        Currently supported values are:

          ``'cobyla'``:
            Uses scipy's fmin_cobyla function
          ``'slsqp'``:
            Uses scipy's fmin_slsqp function (default)

      solvercfg (dict, optional): Keyword arguments that are passed to the
        solver. Common options are:

          acc (float)
            Accuracy of the minimization.
          maxiter (int)
            Maximum number of iterations.
          disp (bool)
            Whether to print information about the minimization.
          iprint (int)
            How much information to print.

        Defaults to ``{}``.
      par (dict, optional): Reference values for the model parameters. They are
        used to determine a metric on observable space (by calling
        model.getQuadraticForm). For parameters not listed in `par` the values
        from ``model.getDefaultParameters()`` are used. Defaults to the empty
        dict.
      obs (dict or None, optional): Specifies a reference point in observable
        space.  This is only relevant if you want to evolve the population to a
        homogenous distribution in observable space using ``evolve``.  Regions
        that are far away from `obs` (depending on the value of `temperature`)
        will be avoided. Defaults to ``None``, in which case ``model.obs`` is
        used.
      neighbours (int, optional): The number of nearest neighbours used to
        estimate the local point density in observable space. Defaults to 2.
      temperature (float, optional): Defaults to 1.0. Set this to a larger value
        to allow ``evolve`` to explore a larger volume in observable space.
      penalty (float, optional): Penalty factor for points that violate the
        non-linear constraints specified in model. Defaults to 1.0.

    Attributes:
      model (Model instance): Associated ``Model`` instance.
      tries (int): Same as `tries` keyword argument.
      retries (int): Same as `retries` keyword argument.
      cvmargin (float): Same as `cvmargin` keyword argument.
      maxcv (float): Same as `maxcv` keyword argument.
      solver (str): Same as `solver` keyword argument.
      solvercfg (dict): Same as `solvercfg` keyword argument.
      obs (dict): Same as `obs` keyword argument.
      neighbours (int): Same as `neighbours` keyword argument.
      temperature (float): Same as `temperature` keyword argument.
      penalty (float): Same as `penalty` keyword argument.

    """

    def __init__(self, model, tries=1, retries=0, cvmargin=10.0,
                 maxcv=1e-4, solver='slsqp', solvercfg={},
                 par={}, obs=None, neighbours=2, temperature=1.0, penalty=1.0):
        self.model = model
        self.tries = tries
        self.retries = retries
        self.cvmargin = cvmargin
        self.maxcv = maxcv
        self.solver = solver
        self.solvercfg = solvercfg
        self.neighbours = neighbours
        self.temperature = temperature
        self.penalty = penalty
        if obs is None:
            self.obs = model.obs
        else:
            self.obs = obs

        self._vpt = None
        self._len = 0
        self._mincons = numpy.inf
        self._parnames = []
        self.update(par)

    def update(self, par={}):
        """Update information about the sample points.

        Note:
          Call this after changing the ``obs``, ``neighbours``, ``temperature``
          or ``penalty`` attributes or after changing the associated ``Model``
          object.

        Args:
          par (dict, optional): Reference values for the model parameters. They
            are used to determine a metric on observable space (by calling
            model.getQuadraticForm). For observables not listed in `par` the
            values from ``model.getDefaultParameters()`` are used. Defaults to
            the empty dict.

        """
        oldparnames = self._parnames
        self._parnames = self.model.getFreeParameters()
        self._obsnames = self.model.getObservableNames()
        fixedpar = self.model.getFixedParameters()
        q = self.model.getDependentQuantities(par)
        self._metric = _dict_to_matrix(self.model.getQuadraticForm(q),
                                       self._obsnames)
        if self._metric.shape[0] == 0:
            self._metric = None
        else:
            self._metric = bsr_matrix(self._metric)

        if self._vpt is None:
            return
        items = [i for i in self._vpt]
        for item in items:
            p = _vector_to_dict(item.pvec, oldparnames)
            newitem = self._makeScanItem(p)
            item.pvec  = newitem.pvec
            item.xvec  = newitem.xvec
            item.chisq = newitem.chisq
            item.cons  = newitem.cons
            self._mincons = min(self._mincons, item.cons)
        self._build(items)

    def _build(self, items):
        self._len = len(items)
        dim = len(self._parnames)
        if self._metric is None:
            self._vpt = sorted(items, key=lambda item: item.cons)
        else:
            self._vpt = VPTree(items, self.distance)
        for item in self._vpt:
            if self._metric is None or len(items) <= 1:
                item.nndist = 1.0
            else:
                for j, d in self._vpt.find(item, self.neighbours):
                    item.nndist = d
            if item.nndist > 0.0:
                item.fitness = -0.5*(item.chisq + self.penalty*item.cons)/ \
                               self.temperature + log(item.nndist)*dim
            else:
                item.fitness = -sys.float_info.max

    def _makeScanItem(self, p):
        pb = {}
        for k, v in p.iteritems():
            lb = self.model.par[k].lowerBound
            if lb is None:
                lb = -inf
            ub = self.model.par[k].upperBound
            if ub is None:
                ub = +inf
            pb[k] = max(min(v, ub), lb)
        q = self.model.getDependentQuantities(pb)
        x = self.model.getCentralValues(q)
        chisq = self.model.getChiSquare(q, self.obs)
        cons = self.model.getConstraintViolation(q)
        return _ScanItem(pvec  = _dict_to_vector(pb, self._parnames),
                         xvec  = _dict_to_vector(x, self._obsnames),
                         chisq = chisq,
                         cons  = cons)

    def sample(self, n, generator=None):
        """Sample the model's parameter space.

        Args:
          n (int): Number of points to sample
          generator (callable or None, optional): Function without arguments
            which generates a random point in parameter space. Defaults to
            ``None``, in which case `model.generateParameterPoint` is used.

        """
        if generator is None:
            generator = self.model.generateParameterPoint
        if len(self.model.getFreeParameters()) == 0:
            n = 1
        if self._vpt is not None:
            items = [i for i in self._vpt]
        else:
            items = []
        for i in range(n):
            p = generator()
            items.append(self._makeScanItem(p))
        self._build(items)

    def clear(self):
        """Clear the data collected with ``sample`` or ``evolve``.

        """
        self._vpt = None
        self._len = 0

    def pariter(self):
        """Return an iterator over the stored points in parameter space.

        """
        if self._vpt is None:
            raise StopIteration()
        else:
            for p in self._vpt:
                yield _vector_to_dict(p.pvec, self._parnames)

    def obsiter(self):
        """Return an iterator over the stored points in observable space.

        """
        if self._vpt is None:
            raise StopIteration()
        else:
            for p in self._vpt:
                yield _vector_to_dict(p.xvec, self._obsnames)

    def __iter__(self):
        """Return an iterator over the stored points.

        Yields:
          par (dict): Model parameters
          obs (dict): Observables
          chisq (float): chi-square value
          cons (float): Constraint violation.

        """
        if self._vpt is None:
            raise StopIteration()
        else:
            for p in self._vpt:
                yield (_vector_to_dict(p.pvec, self._parnames),
                       _vector_to_dict(p.xvec, self._obsnames),
                       p.chisq, p.cons)

    def __len__(self):
        return self._len

    def distance(self, x, y):
        """Compute distance between two points in observable space.

        Args:
          x, y (dict): Points in observable space.

        Returns:
          (float): Distance between `x` and `y`.

        """
        if self._metric is None:
            return 0.0
        d = x.xvec - y.xvec
        dsq = vdot(d, self._metric.dot(d))
        if dsq < 0.0:
            raise MetricError('Non-positive-definite metric encountered')
        return sqrt(dsq)

    def find(self, x):
        """Iterate over collected points, starting with points nearest to `x`.

        Args:
          x (dict): A point in observable space.
          
        Yields:
          p (dict): Parameter values
          x (dict): Observable values
          chisq (float): Chi-square value
          cons (float): Sum of squared constraint functions.

        """
        if self._vpt is None:
            raise StopIteration()
        if self._metric is None:
            for item in self._vpt:
                yield \
                    _vector_to_dict(item.pvec, self._parnames), \
                    _vector_to_dict(item.xvec, self._obsnames), \
                    item.chisq, \
                    item.cons
        else:
            xvec = _dict_to_vector(x, self._obsnames)
            for item, dist in self._vpt.find(_ScanItem(xvec=xvec)):
                yield \
                    _vector_to_dict(item.pvec, self._parnames), \
                    _vector_to_dict(item.xvec, self._obsnames), \
                    item.chisq, \
                    item.cons

    def nearest(self, x):
        """Find nearest neighbour to a point in observable space.

        Args:
          x (dict): Point in observable space.

        Returns (dict):
          Location of a point p in parameter space which is nearest to `x`
          in observable space.

        """
        if self._vpt is None:
            raise ScanDataError('No data.')
        if self._metric is None:
            res = self._vpt[:1]
        else:
            xvec = _dict_to_vector(x, self._obsnames)
            res = [i for i in self._vpt.find(_ScanItem(xvec=xvec), 1)]
        if res:
            return _vector_to_dict(res[0].pvec, self._parnames)
        else:
            return None

    def nnearest(self, x, n):
        """Find n nearest neighbours to a point in observable space.

        Args:
          x (dict): Point in observable space.
          n (int): Number of points to find.

        Returns (list of dict):
          List of dictionaries which specify the locations of (at most) `n`
          points p in parameter space which are closest to `x` in observable
          space.

        """
        if self._vpt is None:
            raise ScanDataError('No data.')
        if self._metric is None:
            res = self._vpt[:n]
        else:
            xvec = _dict_to_vector(x, self._obsnames)
            res = [i for i in self._vpt.find(_ScanItem(xvec=xvec), n)]
        return [_vector_to_dict(i[0].pvec, self._parnames) for i in res]

    def evolve(self, grow, repeat=1, shrink=None, candidates=10, adapt=1.0):
        """Evolve current population towards a more homogenous distribution in
        observable space.

        Args:
          grow (int): Number of points to add to the population.
          repeat (int, optional): Setting `repeat` to `n` is equivalent to
            calling evolve `n` times with the same arguments (and ``repeat=1``).
            Defaults to 1.
          shrink (int or None, optional): Number of points to remove from the
            population.  Defaults to ``None``, in which case `grow` points are
            removed.
          candidates (int, optional): Number of extra candidates to consider for
            reproduction.  Defaults to 10.
          adapt (float, optional): Level of adaptation. Large values mean faster
            adaptation. A value of 0 means no adaptation. Defaults to 1.

        """
        if shrink is None:
            shrink = grow

        for rep in range(repeat):
            dim = len(self._parnames)
            items = SortedCollection(self._vpt, key=lambda x: x.fitness)
            newitems = []
            maxfitness = items[-1].fitness
            for i in xrange(grow):
                i1 = items[int(random.uniform(0, 1)**(1.0/(1.0+adapt))\
                                   *len(items))]

                if self._metric is None:
                    j1 = items[int(random.uniform(0, 1)**(1.0/(1.0+adapt))*\
                                       len(items))]
                    j2 = items[int(random.uniform(0, 1)**(1.0/(1.0+adapt))*\
                                       len(items))]
                else:
                    cand = [j for j, d in self._vpt.find(i1,
                                                         dim + candidates + 1)]
                    del cand[0]
                    j1, j2 = _multichoice(cand, 2)

                pvec = i1.pvec + random.uniform(0, 1)*(j1.pvec - j2.pvec)
                p = _vector_to_dict(pvec, self._parnames)
                newitems.append(self._makeScanItem(p))

            items = [items[i] for i in xrange(shrink, len(items))]
            newitems = newitems + items
            random.shuffle(newitems)
            self._build(newitems)

    def fit(self, obs=None, start=[], tries=None, retries=None, cvmargin=None,
            maxcv=None, solver=None, solvercfg=None, **kwargs):
        """Find the global minimum of a model's chi-square function.

        Args:
          obs (dict or None, optional): Holds the measured values of the
            observables. Defaults to ``None``, in which case ``self.model.obs``
            is used.
          start (list of dicts, optional): Extra starting points to try. 
          tries (int or None, optional): The number of local minimizations to
            try. Defaults to ``None``, in which case the the value of
            :py:attr:`Fitter.tries` is used.
          retries (int or None, optional): Maximum number of retries (with a
            different starting point) after a failed minimization. Defaults to
            ``None``, in which case the value of :py:attr:`Fitter.retries` is
            used.
          cvmargin (float or None, optional): The acceptable amount of constraint
            violation when selecting starting points for the optimization. The
            sum of squares of the constraint functions must be smaller than
            `cvmargin`. Defaults to ``None``, in which case the value of
            :py:attr:`Fitter.cvmargin` is used.
          maxcv (float or None, optional): Maximal value of the constraint
            violation (sum of squares of all violated constraint functions) for
            which the optimization is still considered successful. Defaults to
            ``None``, in which case the value of :py:attr:`Fitter.maxcv` is
            used.
          solver (str or None, optional): Identifies the optimization algorithm
 
            to use.  Currently supported values are:
 
              ``'cobyla'``:
                Uses scipy's fmin_cobyla function
              ``'slsqp'``:
                Uses scipy's fmin_slsqp function

            Defaults to ``None``, in which case the value of
            :py:attr:`Fitter.solver` is used.

          solvercfg (dict, optional): Keyword arguments that are passed to the
            solver. Common options are:

              acc (float): Accuracy of the minimization.
              maxiter (int): Maximum number of iterations.
              disp (bool): Whether to print information about the minimization.
              iprint (int): How much information to print.

            Defaults to ``None``, in which case the value of
            :py:attr:`Fitter.solvercfg` is used.
          **kwargs (optional): All other keyword arguments are passed to the
            ``fit`` function, which does the local optimization.

        Returns:
          A ``FitResult`` instance with the result of the successful local
          minimization that found the smallest chi-square.

        """
        if tries is None:
            tries = self.tries
        if retries is None:
            retries = self.retries
        if cvmargin is None:
            cvmargin = self.cvmargin
        if maxcv is None:
            maxcv = self.maxcv
        if solver is None:
            solver = self.solver
        if obs is None:
            obs = self.model.obs

        cfg = self.solvercfg.copy()
        if solvercfg is not None:
            cfg.update(solvercfg)

        startingpoints = start[:]
        tries = max(tries, 0)
        retries = max(retries, 0)
        if tries > 0:
            for p, x, chisq, cons in self.find(obs):
                if cons < self._mincons + cvmargin:
                    startingpoints.append(p)
                if len(startingpoints) >= len(start)+tries+retries:
                    break
        return fit(self.model, obs=obs, start=startingpoints,
                   tries=len(start)+tries, maxcv=maxcv,
                   solver=solver, solvercfg=cfg, **kwargs)

    


